import argparse
import random

from io import TextIOWrapper
from networkx import random_degree_sequence_graph, is_graphical, shortest_path, has_path
from pathlib import Path
from typing import List, Tuple

def _parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('-o', '--output', type=Path, required=True, help='Output directory')
    parser.add_argument('--problems', type=int, required=True, help='Number of problems to generate')
    parser.add_argument('--min_size', type=int, required=True, help='Minimum size of graph')
    parser.add_argument('--max_size', type=int, required=True, help='Maximum size of graph')
    parser.add_argument('--min_robots', type=int, required=True, help='Minimum number of robots')
    parser.add_argument('--max_robots', type=int, required=True, help='Maximum number of robots')
    parser.add_argument('--different_graphs', action='store_true', help='If robots navigate differently')
    args = parser.parse_args()
    return args


def _write_domain(stream: TextIOWrapper):
    stream.write('(define (domain robot-vacuum)\n')
    stream.write('  (:requirements :strips :negative-preconditions)\n')
    stream.write('  (:predicates (at ?r ?l) (dirty ?l) (adjacent ?r ?l1 ?l2))\n')
    stream.write('\n')
    stream.write('  (:action drive\n')
    stream.write('   :parameters (?r ?l1 ?l2)\n')
    stream.write('   :precondition (and (at ?r ?l1) (adjacent ?r ?l1 ?l2))\n')
    stream.write('   :effect (and (not (at ?r ?l1)) (at ?r ?l2)))\n')
    stream.write('\n')
    stream.write('  (:action clean\n')
    stream.write('   :parameters (?r ?l)\n')
    stream.write('   :precondition (and (dirty ?l) (at ?r ?l))\n')
    stream.write('   :effect (and (not (dirty ?l))))\n')
    stream.write(')\n')


def _write_problem(stream: TextIOWrapper, size: int, dirty_index: int, robots: List[Tuple[List[Tuple[int, int]], List[int], List[int]]]):
    starting_offsets = [random.randint(0, len(x) - 1) for (_, x, _) in robots]
    stream.write('; Summary of the generated problem:\n')
    stream.write('; Dirt: {}\n'.format(dirty_index))
    stream.write('; Robot [Start, Distance]: Edges\n')
    for (robot_index, (robot_edges, start_indices, start_distances)) in enumerate(robots):
        start_index = start_indices[starting_offsets[robot_index]]
        start_distance = start_distances[starting_offsets[robot_index]]
        stream.write('; * rob_{} [{}, {}]:'.format(robot_index, start_index, start_distance))
        for edge in robot_edges:
            stream.write(' ({}, {})'.format(*edge))
        stream.write('\n')
    stream.write('\n')
    stream.write('(define (problem robot-vacuum)\n')
    stream.write('  (:domain robot-vacuum)\n')
    stream.write('  (:objects')
    for robot_index, _ in enumerate(robots):
        stream.write(' rob_{}'.format(robot_index))
    for location_index in range(size):
        stream.write(' loc_{}'.format(location_index))
    stream.write(')\n')
    stream.write('  (:init\n')
    stream.write('    (dirty loc_{})\n'.format(dirty_index))
    for robot_index, (robot_edges, start_indices, _) in enumerate(robots):
        stream.write('    (at rob_{} loc_{})\n'.format(robot_index, start_indices[starting_offsets[robot_index]]))
        for edge in robot_edges:
            stream.write('    (adjacent rob_{} loc_{} loc_{})\n'.format(robot_index, edge[0], edge[1]))
            stream.write('    (adjacent rob_{} loc_{} loc_{})\n'.format(robot_index, edge[1], edge[0]))
    stream.write('  )\n')
    stream.write('  (:goal (and (not (dirty loc_{}))))\n'.format(dirty_index))
    stream.write(')\n')


def frange(start: int, end: int, count: int) -> List[float]:
    result = []
    step = (end - start) / count
    for _ in range(count):
        result.append(start)
        start += step
    return result


def _generate_graph(graph_size: int, distance_diff: int, seed: int) -> Tuple[List[Tuple[int, int]], List[int], List[int]]:

    def generate_graphical_degree_sequence() -> List[int]:
        degrees = [random.randint(2, 3) for _ in range(graph_size)]
        while not is_graphical(degrees):
            for idx, val in enumerate(degrees):
                if val > 1:
                    degrees[idx] = val - 1
                    break
        return degrees

    def get_starting_indices(distances: List[int]) -> Tuple[List[int], int]:
        indices = list(zip(distances, range(0, len(distances))))
        indices.sort()
        max_distance = indices[-1][0]
        min_distance = max(1, max_distance - distance_diff)
        for index, (distance, _) in enumerate(indices):
            if distance >= min_distance:
                starting_indices = [starting_index for _, starting_index in indices[index:]]
                starting_distances = [starting_distance for starting_distance, _ in indices[index:]]
                return starting_indices, starting_distances
        raise Exception("This should never happen")

    random.seed(seed)
    degrees = generate_graphical_degree_sequence()
    graph = random_degree_sequence_graph(degrees, seed)
    shortest_paths = [len(shortest_path(graph, 0, i)) - 1 if has_path(graph, 0, i) else -1 for i in graph.nodes]
    start_indices, start_distances = get_starting_indices(shortest_paths)
    return (list(graph.edges), start_indices, start_distances)


if __name__ == "__main__":
    args = _parse_arguments()
    args.output.mkdir(parents=True, exist_ok=True)
    domain_path = args.output / 'domain.pddl'
    with open(domain_path, 'w') as domain_stream:
        _write_domain(domain_stream)
    print('Generated {}'.format(str(domain_path)))

    def settings(n, m, i): return [int(round(x)) for x in frange(n, m, i)]
    sizes = settings(args.min_size, args.max_size, args.problems)
    robots = settings(args.min_robots, args.max_robots, args.problems)

    for i in range(args.problems):
        graph_size = sizes[i]
        num_robots = robots[i]
        if args.different_graphs: graphs = [_generate_graph(graph_size, 0, i + n) for n in range(num_robots)]
        else: graphs = [_generate_graph(graph_size, 3, i)] * num_robots
        plan_length = min([min(distances) for (_, _, distances) in graphs])
        problem_path = args.output / 'p{:0>2}_nodes-{}_robots-{}_plan-{}.pddl'.format(i + 1, graph_size, num_robots, plan_length + 1)
        with open(problem_path, 'w') as problem_stream:
            _write_problem(problem_stream, graph_size, 0, graphs)
        print('Generated {}'.format(str(problem_path)))
